from .frames import frame_manager
from .parsing import split_variables, unquote

# TODO, rewrite this part, the return True everywhere is bad... or differentiate between validators and probes?
# or always return, or always rise.


def get_dotted(ipython_globals, param):
    parts = param.split('.')
    try:
        # e.g. "import os.path", param="os.path"
        return ipython_globals[param]
    except KeyError:
        value = ipython_globals[parts[0]]
        for part in parts[1:]:
            value = getattr(value, part)
        return value


class ParametersValidator:

    # TODO rename module to folder (or other way of locating something in the vault)?
    def module(self, param: str):
        """Valid name of a folder"""
        return unquote(param) == param

    def path(self, param: str):
        """Valid path in the storage represented as a string; has to be packed in (single or double) quotes."""
        return unquote(param) != param

    def _get_from_globals(self, param: str, kind: str):
        # 1. is a valid identifier?
        parts = param.split('.')
        if not all([part.isidentifier() for part in parts]):
            raise ValueError(f"'{param}' is not a valid {kind} name")
        # 2. exists in the global namespace?
        ipython_globals = frame_manager.get_ipython_globals()
        try:
            return get_dotted(ipython_globals, param)
        except KeyError:
            raise NameError(f"{kind} '{param}' is not defined in the global namespace")

    def one_variable(self, param: str):
        """Valid Python variable"""
        self._get_from_globals(param, 'variable')
        return True

    def one_or_many_variables(self, param: str):
        """Valid Python variables separated by colon"""
        for v in split_variables(param):
            self._get_from_globals(v, 'variable')
        return True

    def function(self, param: str):
        """Valid Python function"""
        function = self._get_from_globals(param, 'function')
        if not callable(function):
            raise ValueError(f'{function} is not a function')
        return True

    def valid_id(self, param: str):
        """Valid Python identifier"""
        assert param.isidentifier()
        return True

    def one_or_many_valid_id(self, param: str):
        for v in split_variables(param):
            assert v.isidentifier()
        return True

    def hash_method(self, param: str):
        """Hash method, one of CRC32 or SHA256"""
        assert param in {'CRC32', 'SHA256'}
        return True

    def hash(self, param: str):
        """Hash from CRC32 or SHA256"""
        if len(param) == 8 or len(param) == 64:
            return True
        else:
            raise ValueError('Hash needs to have either 8 (CRC32) or 64 (SHA256) characters')



import pandas as pd
import numpy as np


def mb(x):
    """Represent as MB"""
    return x / 10**6


def optimize_memory(
    df: pd.DataFrame,
    categorical_threshold=0.2,
    categorise_numbers=False,
    categorise_booleans=False,
    inplace=True,
    report=True
):

    before = df.memory_usage(index=True, deep=True).sum()

    if not inplace:
        df = df.copy()

    for column in df.columns:
        s = df[column]

        if s.dtype.name == 'category':
            continue

        if np.issubdtype(s.dtype, np.integer):
            if s.min() >= -125 and s.max() < 125 and not s.isnull().any():
                df[column] = s.astype('int8')
                continue

        # attempt to categorise
        values = set(s)

        if np.issubdtype(s.dtype, np.number) and not categorise_numbers:
            continue

        if np.issubdtype(s.dtype, np.bool_) and not categorise_booleans:
            continue

        if len(values) / len(s) < categorical_threshold:
            try:
                sorted_categories = sorted([v for v in values if not pd.isnull(v)])
                ordered = True
            except TypeError as e:
                sorted_categories = None
                ordered = False
                print(f'Not sorting categories for {column}: {e}')
            df[column] = pd.Categorical(s, categories=sorted_categories, ordered=ordered)

    after = df.memory_usage(deep=True).sum()

    if report:
        percent = (before - after) / before * 100
        print(
            f'Reduced memory usage by {percent:.2f}%,'
            f' from {mb(before):.2f} MB to {mb(after):.2f} MB.'
        )

    return df



class FrameManager:
    """Just to make unit testing easier"""

    def __init__(self, ipython_globals=None):
        self.ipython_globals = ipython_globals

    def get_ipython_globals(self):
        assert self.ipython_globals
        return self.ipython_globals


frame_manager = FrameManager()



# TODO
from typing import Dict, Callable

from .vault import Vault


class DynamicVault:

    def __init__(self, path, vault: Vault, cache=True):
        self.path = path
        self.vault = vault
        self.importers = {}
        self.cache = cache
        self.cached = {}

    def set_importers(self, importers: Dict[str, Callable]):
        self.importers = importers

    def __getattr__(self, key):
        if key in self.__dir__():
            if self.cache and key in self.cached:
                return self.cached[key]
            importer = self.importers.get(key, None)
            # TODO: display metadata
            value = self.vault.load_object(self.path + '/' + key, key, importer, to_globals=False)
            if self.cache:
                self.cached[key] = value
            return value
        raise AttributeError

    def __dir__(self):
        """Make members tab-completable"""
        return self.vault.list_members(relative_to=self.path)



import gzip
import json
from typing import List
from warnings import warn
from datetime import datetime

from IPython.display import display, Markdown
from IPython import get_ipython
from IPython.core.magic import Magics, magics_class, line_magic, needs_local_scope

from .action import Action
from .actions import StoreAction, ImportAction, DeleteAction, AssertAction
from .frames import frame_manager
from .parsing import parse_arguments, clean_line
from .vault import Vault


def one(values):
    assert len(values) == 1
    return list(values)[0]


@magics_class
class VaultMagics(Magics):
    """The `%vault` magic provides a reproducible caching mechanism for variables exchange between notebooks.

    To open the vault use `%open_vault` magic.
    """

    defaults = {
        'path': 'storage.zip',
        'encryption_variable': None,
        'secure': True,
        'optimize_df': True,
        'timestamp': True,
        'metadata': True,
        'logs_path': '{path}.vault.log.gz',
        'gzip_logs': True,
        'report_memory_gain': False,
        # aggressive memory optimisation by categorising numbers
        'numbers_as_categories': False,
        'booleans_as_categories': False
        # 'allowed_duration': 30,  # seconds
    }

    assert len(set([d[0] for d in defaults])) == len(defaults)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.settings = None
        self.current_vault: Vault = None

    actions: List[Action] = [
        StoreAction,
        ImportAction,
        DeleteAction,
        AssertAction
    ]

    @needs_local_scope
    @line_magic
    def open_vault(self, line, local_ns=None):
        """Open a zip archive for the vault. Once opened, all subsequent `%vault` magics operate on this archive."""
        frame_manager.ipython_globals = local_ns
        self.settings = parse_arguments(line, self.defaults)
        self.current_vault = Vault(self.settings)
        if self.settings['secure'] and not self.settings['encryption_variable']:
            warn(
                'Encryption variable not set - no encryption will be used.'
                ' Your data may be susceptible,'
                ' and you may not be able to access stored objects if those were previously encrypted.'
                ' Please provide the name of the environment variable with the storage key using `-e env_var_name`, or'
                ' set `--secure False` to silence this warning if you do not need additional protection.'
            )

    open_vault.__doc__ += '\n\nOpen vault arguments:\n\n' + '\n'.join([
        f'\t --{key}, -{key[0]}, default {value}'
        for key, value in defaults.items()
    ])

    def _ensure_configured(self):
        if not self.settings:
            raise Exception('Please setup the storage with %open_vault first.')

    def append_to_logs(self, metadata):
        logs_path = self.settings['logs_path'].format(**self.settings)
        opener = (gzip.open if self.settings['gzip_logs'] else open)
        with opener(logs_path, mode='ta+') as f:
            f.write(json.dumps(metadata) + '\n')

    def extract_arguments(self, line):
        iterable = iter(clean_line(line))
        return {key: next(iterable) for key in iterable}

    def select_action(self, arguments):

        actions = {
            action.main_keyword: action
            for action in self.actions
        }

        requested_actions = set(actions).intersection(arguments)
        requested_action = one(requested_actions)

        action_class = actions[requested_action]
        action = action_class(vault=self.current_vault)
        return action

    @needs_local_scope
    @line_magic
    def vault(self, line, local_ns=None):
        """Perform one of the available actions, print the description and save metadata in the cell."""
        self._ensure_configured()
        frame_manager.ipython_globals = local_ns

        started = self._timestamp()

        arguments = self.extract_arguments(line)
        action = self.select_action(arguments)
        metadata = action.perform(arguments)

        finished = self._timestamp()

        # if finished - started > settings['allowed_duration']:
        # warn that the operations took longer than expected

        metadata['started'] = started.isoformat()
        metadata['finished'] = finished.isoformat()
        metadata['finished_human_readable'] = finished.strftime('%A, %d. %b %Y %H:%M')
        metadata['command'] = line

        self.append_to_logs(metadata)

        display(Markdown(
            (
                action.short_stamp(metadata)
                if self.settings['timestamp'] else
                None
            ),
            metadata=(
                metadata
                if self.settings['metadata'] else
                None
            )
        ))

    vault.__doc__ += '\n\nVault commands:\n\n' + '\n'.join(
        [action.explain() for action in actions]
    )

    @staticmethod
    def _timestamp():
        return datetime.utcnow()


ip = get_ipython()
if ip:
    ip.register_magics(VaultMagics)
else:
    warn('Could not register VaultMagics - are you running from IPython?')



from typing import Dict, Callable, List, Union
from abc import ABC, abstractproperty
from collections import Counter

from .vault import Vault
from .frames import frame_manager
from .parameters import get_dotted


Metadata = Dict[str, Union[str, List[Dict]]]


class Syntax:

    def __init__(self, required, optional=None, disallowed=None):
        self.required = required
        self.optional = optional or {}
        self.disallowed = disallowed or {}

    def calc_concordance(self, arguments, target, raise_exceptions=False):
        concordance = 0

        ruleset = getattr(self, target)
        misses = {}

        for arg, validator in ruleset.items():

            key = arg + ' <' + validator.__name__ + '>'

            if arg not in arguments:
                misses[key] = 'is missing'
                continue

            concordance += 0.5
            value = arguments[arg]

            try:
                if validator(value):
                    concordance += 0.5
                else:
                    misses[key] = 'not a ' + validator.__name__
            except Exception as e:
                if raise_exceptions:
                    raise e
                misses[key] = e

        return {
            'ratio': concordance / len(ruleset) if len(ruleset) else 0,
            'misses': misses
        }

    def diff(self, arguments):
        return ', '.join([
            f"'{keyword}': {explanation}"
            for keyword, explanation in
            self.calc_concordance(arguments, 'required', raise_exceptions=False)['misses'].items()
        ])

    def validate(self, arguments):
        assert self.calc_concordance(arguments, 'required', raise_exceptions=True)['ratio'] == 1
        # if it does not raise, it's ok
        self.calc_concordance(arguments, 'optional', raise_exceptions=True)
        self.check_disallowed(arguments)

    def check_disallowed(self, arguments):
        for keyword, error in self.disallowed.items():
            if keyword in arguments:
                raise ValueError(error)

    def _repr_args(self, args):
        return ' '.join([
            f'{keyword} <{param.__name__}>'
            for keyword, param in args.items()
        ])

    def __repr__(self):
        return (
            self._repr_args(self.required)
            + ' ['
            + self._repr_args(self.optional)
            + ']'
        )


class Action(ABC):

    def __init__(self, vault: Vault):
        self.vault: Vault = vault

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        cls.__doc__ = (cls.__doc__ or '') + cls.explain()

    @abstractproperty
    def handlers(self) -> Dict[Callable, Syntax]:
        """Should be ordered from the most specific to the most general."""

    @abstractproperty
    def verb(self) -> str:
        """Past form of a verb describing the finished action."""

    @abstractproperty
    def main_keyword(self) -> str:
        """The main keyword which defines the kind of action"""

    def choose_handler(self, arguments):
        for handler, syntax in self.handlers.items():
            concordance = syntax.calc_concordance(arguments, 'required', raise_exceptions=False)['ratio']

            if concordance == 1:
                return handler

        return None

    def closest_syntax(self, arguments, n=3):
        counter = Counter({
            syntax: syntax.calc_concordance(arguments, 'required', raise_exceptions=False)['ratio']
            for handler, syntax in self.handlers.items()
        })
        return counter.most_common(n)

    def syntax_help(self, n=None, arguments=None):
        arguments = arguments or {}
        closest = [
            str(syntax) + (
                '\n\t\t> ' + syntax.diff(arguments)
                if arguments else
                ''
            )
            for syntax, concordance in self.closest_syntax(arguments, n=n)
        ]
        return '\n\t - '.join(map(str, [''] + closest))

    def perform(self, arguments) -> Metadata:
        # choose handler using syntax concordance with the required arguments
        handler = self.choose_handler(arguments)
        if not handler:
            error = 'No command matched. Did you mean:' + self.syntax_help(n=3, arguments=arguments)
            raise ValueError(error)

        # validate the optional and disallowed parts of the syntax
        syntax = self.handlers[handler]
        syntax.validate(arguments)

        bound_handler = getattr(self, handler.__name__)

        return {
            'action': self.main_keyword,
            'result': bound_handler(arguments)
        }

    def short_stamp(self, metadata: Metadata) -> str:
        """Return short, human readable description of the action"""

        def repr_result(result, hash_method='crc32'):
            hashcodes = [
                f'{result[file][hash_method]}'
                for file in ['old_file', 'new_file']
                if file in result
            ]
            return f"`{result['subject']}`" + ' (' + ' → '.join(hashcodes) + ')'

        results_n = len(metadata['result'])

        return (
            self.verb.capitalize()
            + (':\n\n - ' if results_n > 1 else ' ')
            + '\n - '.join([
                repr_result(result)
                for result in metadata['result']
            ])
            + ('\n\n' if results_n > 1 else ' ')
            + 'at '
            + metadata['finished_human_readable']
        )

    @property
    def ipython_globals(self):
        return frame_manager.get_ipython_globals()

    def with_function(self, arguments):
        if 'with' in arguments:
            func_name = arguments['with']
            return get_dotted(self.ipython_globals, func_name)
        return None

    @classmethod
    def explain(cls):
        instance = cls(None)
        return f'# {instance.main_keyword}' + instance.syntax_help()



def bool_or_str(value: str):
    if value == 'True':
        return True
    if value == 'False':
        return False
    return value


def clean_line(line: str) -> str:
    """Split/tokenize a line with keywords and parameters.

    TODO: rewrite using grammar or lexer?
    """
    pieces = ['']
    quote = ''
    prev = ''
    for c in line.strip():
        if quote:
            if prev != '\\' and c == quote:
                quote = ''
                pieces[-1] += c
                pieces.append('')
            else:
                pieces[-1] += c
        else:
            if c == '"' or c == "'":
                assert pieces[-1] == ''
                pieces[-1] += c
                quote = c
            elif c.isspace():
                if pieces[-1] != '' and pieces[-1][-1] != ',':
                    pieces.append('')
            elif c == '#':
                break
            else:
                pieces[-1] += c
        prev = c

    if pieces[-1] == '':
        pieces = pieces[:-1]

    return pieces


def parse_arguments(line, defaults):
    iterable = iter(clean_line(line))
    user_arguments = {key.lstrip('-'): next(iterable) for key in iterable}

    config = defaults.copy()
    for key in config:
        if key in user_arguments:
            config[key] = bool_or_str(user_arguments[key])
        else:
            short_key = key[0]
            if short_key in user_arguments:
                config[key] = bool_or_str(user_arguments[short_key])
    return config


def unquote(text):
    if text.startswith('"'):
        assert text.endswith('"')
        text = text[1:-1].replace(r'\"', '"')

    if text.startswith("'"):
        assert text.endswith("'")
        text = text[1:-1].replace(r"\'", "'")

    return text


def split_variables(variables: str):
    variables = [
        v.strip() for v in variables.split(',')
    ]
    for v in variables:
        assert v != '*'
    assert len(variables) >= 1
    return variables



from tempfile import NamedTemporaryFile
from io import BytesIO
import os
import hashlib
from typing import Dict
import zlib
from warnings import warn

from pandas import read_csv

from .seven_zip import SevenZip
from .memory import optimize_memory
from .frames import frame_manager


class Vault:

    def __init__(self, settings: Dict):
        self.settings = settings

    def list_members(self, relative_to=None):
        return self.archive.list_members(relative_to=relative_to)

    def _default_exporter(self, variable, file_object):
        # line terminator set to '\n' to have the same hashes between Unix and Windows
        return variable.to_csv(file_object, sep='\t', line_terminator='\n')

    @property
    def archive(self):
        return SevenZip(
            archive_path=self.settings['path'],
            password=self._password
        )

    def save_object(self, path, value, exporter, **metadata):

        if not exporter:
            exporter = self._default_exporter

        old_checksum_crc, old_checksum_sha = None, None

        archive = self.archive

        if path in archive:
            old_checksum_crc = archive.calc_checksum(path, method='CRC32')
            old_checksum_sha = archive.calc_checksum(path, method='SHA256')

        with NamedTemporaryFile(delete=False) as f:

            try:
                f.close()
                try:
                    exporter(value, f.name)
                except AttributeError as e:
                    # json
                    if str(e) != "'str' object has no attribute 'write'":
                        raise
                    with open(f.name, 'w') as f2:
                        exporter(value, f2)
                except TypeError as e:
                    # pickle
                    if str(e) != "file must have a 'write' attribute":
                        raise
                    with open(f.name, 'wb') as f2:
                        exporter(value, f2)
                archive.add_file(f.name, rename=path)
            except Exception as e:
                raise e
            finally:
                os.remove(f.name)

        new_checksum_crc = archive.calc_checksum(path, method='CRC32')
        new_checksum_sha = archive.calc_checksum(path, method='SHA256')
        archive.check_integrity()

        return {
            'new_file': {
                'crc32': new_checksum_crc,
                'sha256': new_checksum_sha
            },
            'old_file': {
                'crc32': old_checksum_crc,
                'sha256': old_checksum_sha
            },
            **metadata
        }

    @property
    def _password(self):
        key_name = self.settings['encryption_variable']

        return (
            os.environ[key_name]
            if key_name else
            False
        )

    def _default_importer(self, file_object):
        df = read_csv(file_object, sep='\t', index_col=0, parse_dates=True)
        if self.settings['optimize_df']:
            df = optimize_memory(
                df,
                categorise_numbers=self.settings['numbers_as_categories'],
                categorise_booleans=self.settings['booleans_as_categories'],
                report=self.settings['report_memory_gain']
            )
        return df

    def load_object(self, path, variable_name, importer=None, to_globals=True):
        if not importer:
            importer = self._default_importer

        archive = self.archive
        info = archive.get_info(path)

        with archive.open(path) as f:
            content = f.read()
            crc_as_int = zlib.crc32(content)
            new_checksum_crc = hex(crc_as_int)[2:].upper()
            new_checksum_sha = hashlib.sha256(content).hexdigest().upper()
            obj = importer(BytesIO(content))
            if to_globals:
                frame_manager.get_ipython_globals()[variable_name] = obj

        # check integrity of a single file
        if info.CRC != crc_as_int:
            if info.CRC == 0:
                # https://sourceforge.net/p/sevenzip/discussion/45798/thread/c284a85f3f/
                warn(
                    'CRC not found, cannot verify integrity (note:'
                    ' this is expected for newer versions of 7zip when using AES encryption)'
                )
            else:
                raise ValueError(f'CRC do not match: {info.CRC} { crc_as_int}')

        metadata = {
            'new_file': {
                'crc32': new_checksum_crc,
                'sha256': new_checksum_sha
            },
            'subject': variable_name
        }

        if to_globals:
            return metadata
        else:
            return obj

    def check_integrity(self):
        self.archive.check_integrity()

    def remove_object(self, path):

        archive = self.archive

        old_checksum_crc = archive.calc_checksum(path, method='CRC32')
        old_checksum_sha = archive.calc_checksum(path, method='SHA256')

        archive.delete(path)

        return [{
            'old_file': {
                'crc32': old_checksum_crc,
                'sha256': old_checksum_sha
            },
            'subject': path
        }]



from .action import Action, Syntax
from .parameters import ParametersValidator
from .parsing import split_variables, unquote
from .dynamic_vault import DynamicVault

params = ParametersValidator()


class StoreAction(Action):
    main_keyword = 'store'
    verb = 'stored'

    def store_in_module_as(self, arguments):
        path = arguments['in'] + '/' + arguments['as']
        variable = arguments['store']
        return self._store({path: variable}, arguments)

    def store_in_module(self, arguments):
        return self._store(
            {
                arguments['in'] + '/' + variable: variable
                for variable in split_variables(arguments['store'])
            },
            arguments
        )

    def store_in_path(self, arguments):
        path = unquote(arguments['in'])
        variable = arguments['store']
        return self._store({path: variable}, arguments)

    handlers = {
        store_in_module_as: Syntax(
            required={'store': params.one_or_many_variables, 'in': params.module, 'as': params.valid_id},
            optional={'with': params.function}
        ),
        store_in_module: Syntax(
            required={'store': params.one_or_many_variables, 'in': params.module},
            optional={'with': params.function, 'as': params.valid_id}
        ),
        store_in_path: Syntax(
            required={'store': params.one_variable, 'in': params.path},
            optional={'with': params.function},
            disallowed={
                'as': (
                    '"as" is not allowed for storing in path'
                    ' (it would be redundant as the path already specifies the target).'
                    ' If you wanted to point to a module in the archive, remove the quotes around {in}.'
                )
            }
        )
    }

    def _store(self, variables_by_paths, arguments):
        exporter = self.with_function(arguments)

        return [
            self.vault.save_object(path, self.ipython_globals[variable], exporter, subject=variable)
            for path, variable in variables_by_paths.items()
        ]


class ImportAction(Action):
    """Load a variable(s) exported from a notebook arbitrary data from the active storage.

    > %vault from notebook_path import variable
    > %vault from notebook_path import variable with your_function as variable
    > %vault import 'file.tsv' with your_function as variable

    It also allows you to specify custom import function, which:
        - has to be available in the global or local namespace
        - should accept a file object
        - should return the data loaded from the file object
    """

    main_keyword = 'import'
    verb = 'imported'

    def from_module_import_as(self, arguments):
        path = arguments['from']
        variable = arguments['import']
        name = arguments.get('as', variable)

        return self._import(
            {path + '/' + variable: name},
            arguments
        )

    def from_module_import(self, arguments):
        path = arguments['from']
        variables = split_variables(arguments['import'])

        return self._import(
            {
                path + '/' + variable: variable
                for variable in variables
            },
            arguments
        )

    def import_path_as(self, arguments):
        path = arguments['import']

        return self._import(
            {unquote(path): arguments['as']},
            arguments
        )

    def import_module(self, arguments):
        path = arguments['import']
        parent = path.split('.')[0]
        name = arguments.get('as', parent)

        if 'as' not in arguments:
            path = parent

        dynamic_vault = DynamicVault(path=path, vault=self.vault)
        self.ipython_globals[name] = dynamic_vault

        return []

    handlers = {
        from_module_import_as: Syntax(
            required={'import': params.valid_id, 'from': params.module, 'as': params.valid_id},
            optional={'with': params.function}
        ),
        from_module_import: Syntax(
            required={'import': params.one_or_many_valid_id, 'from': params.module},
            optional={'with': params.function, 'as': params.valid_id}
        ),
        import_path_as: Syntax(
            required={'import': params.path, 'as': params.valid_id},
            optional={'with': params.function}
        ),
        import_module: Syntax(
            required={'import': params.module},
            optional={'as': params.valid_id},
            disallowed={
                'with': (
                    '"with" not allowed for module import;'
                    ' to change the loading function for variables in the module,'
                    ' use module.set_importers({variable: function})'
                )
            }
        )
    }

    def _import(self, variables_by_paths, arguments):
        importer = self.with_function(arguments)

        return [
            self.vault.load_object(path, variable, importer)
            for path, variable in variables_by_paths.items()
        ]


class DeleteAction(Action):

    main_keyword = 'del'
    verb = 'deleted'

    def delete_variable(self, arguments):
        location = arguments['del']
        path = arguments['from'] + '/' + location

        return self._delete(arguments, path)

    def delete_path(self, arguments):
        location = arguments['del']
        path = unquote(location)

        return self._delete(arguments, path)

    handlers = {
        delete_variable: Syntax(
            required={'del': params.valid_id, 'from': params.module}
        ),
        delete_path: Syntax(
            required={'del': params.path}
        )
    }

    def _delete(self, arguments, path):
        # TODO test for folders, require all to remove a folder?
        # TODO: what to do about wildcards?
        assert '*' not in path

        return self.vault.remove_object(path)


class AssertAction(Action):
    """Verify the checksum of the input file, raise AssertionError if it differs.

    By default the checksum is calculated with CRC32."""
    main_keyword = 'assert'
    verb = 'verified'

    def assert_variable_hash(self, arguments):
        path = arguments['in'] + '/' + arguments['assert']
        return self._assert_hash(path=path, arguments=arguments)

    def assert_path_hash(self, arguments):
        return self._assert_hash(path=arguments['assert'], arguments=arguments)

    def _assert_hash(self, path, arguments):
        expected = arguments['is']
        method = arguments.get('with', 'CRC32')
        calculated = self.vault.archive.calc_checksum(path, method=method)
        assert expected == calculated
        return [{
            'subject': path,
            'old_file': {
                'crc32': calculated
            }
        }]

    handlers = {
        assert_variable_hash: Syntax(
            required={'assert': params.one_variable, 'in': params.module, 'is': params.hash},
            optional={'with': params.hash_method}
        ),
        assert_path_hash: Syntax(
            required={'assert': params.path, 'is': params.hash},
            optional={'with': params.hash_method}
        )
    }



import subprocess
from contextlib import contextmanager
from pathlib import Path
from zipfile import ZipFile, ZipInfo
from io import BytesIO


class SevenZip:

    command = '7z'

    def __init__(self, archive_path: str, password=None):
        self.path = archive_path
        self.password = password

    @contextmanager
    def open(self, file_path, mode='r', password: str = None, use_7z: bool = True):
        password = password or self.password

        if use_7z:
            yield BytesIO(self._execute(
                'e', file_path, '-so', *self._password_arg(password),
                decode=False, expect_ok_message=False
            ))
        else:
            if password:
                password = password.encode()

            with ZipFile(self.path) as archive:
                yield archive.open(file_path, mode=mode, pwd=password)

    def _execute(self, command, *args, decode: bool = True, expect_ok_message: bool = True):
        o = subprocess.check_output(
            [self.command, command, self.path, *args]
        )
        if expect_ok_message:
            assert b'Everything is Ok' in o
        if decode:
            o = o.decode('utf-8')
        return o

    def rename(self, old_path: str, new_path: str):
        return self._execute('rn', old_path, new_path)

    def exists(self):
        return Path(self.path).exists()

    def __contains__(self, file_path: str):
        if not self.exists():
            return False
        return file_path in self.list_members()

    def _password_arg(self, password: str):
        """Set password argument for given argument:
        - the password set at initialization if `password` is None,
        - given `password` if not None,
        - empty if `password` is `False`
        """
        if password is False:
            return []
        if password is not None:
            return ['-p' + password]
        if self.password:
            return ['-p' + self.password]
        return []

    def calc_checksum(self, file_path: str, method='CRC32', password=None):
        """Method: usually one of CRC32, CRC64, SHA1, SHA256, BLAKE2sp, CRC32 by default.
        """
        args = [file_path, '-scrc' + method] + self._password_arg(password)
        result = self._execute('t', *args)
        lines = [line for line in result.split('\n') if 'for data:' in line]
        assert len(lines) == 1
        method_used, hashsum = lines[0].split('for data:')
        return hashsum.strip()

    def check_integrity(self, password=None):
        return self._execute('t', *self._password_arg(password))

    def _add_file(self, file_path: str, password=None):
        args = ['-y', file_path] + self._password_arg(password)
        return self._execute('a', *args)

    def add_file(self, file_path: str, password=None, rename: str = False):
        assert rename is not True
        if not rename:
            self._add_file(file_path, password=password)
        else:
            added_path_in_archive = Path(file_path).name

            try:
                self._add_file(file_path, password=password)
                if rename in self:
                    self.delete(rename)
                self.rename(added_path_in_archive, rename)
            except Exception:
                self.delete(added_path_in_archive)
                raise

    def delete(self, file_to_remove: str):
        return self._execute('d', file_to_remove)

    def get_info(self, path) -> ZipInfo:
        with ZipFile(self.path) as archive:
            return archive.getinfo(path)

    def list_members(self, relative_to=''):
        with ZipFile(self.path) as archive:
            return [
                (
                    member.split(relative_to + '/')[1]
                    if relative_to else
                    member
                )
                for member in archive.namelist()
                if member.startswith(relative_to + '/' if relative_to else '')
            ]



from pathlib import Path
from setuptools import setup
from setuptools import find_packages


if __name__ == '__main__':
    setup(
        name='data_vault',
        packages=find_packages(),
        version='0.4.5',
        license='MIT',
        description='IPython magic for simple, organized, compressed and encrypted storage of files',
        long_description=Path('README.md').read_text(),
        author='Michal Krassowski',
        author_email='krassowski.michal+pypi@gmail.com',
        url='https://github.com/krassowski/data-vault',
        keywords=['jupyter', 'jupyterlab', 'notebook', 'ipython', 'storage', 'store', 'magic', 'vault'],
        classifiers=[
            'Development Status :: 5 - Production/Stable',
            'License :: OSI Approved :: MIT License',
            'Framework :: IPython',
            'Framework :: Jupyter',
            'Operating System :: Microsoft :: Windows',
            'Operating System :: POSIX :: Linux',
            'Operating System :: MacOS',
            'Topic :: Utilities',
            'Topic :: Database',
            'Topic :: System :: Archiving',
            'Topic :: System :: Archiving :: Compression',
            'Topic :: Software Development :: User Interfaces',
            'Topic :: Software Development :: Libraries :: Python Modules',
            'Intended Audience :: Developers',
            'Intended Audience :: Science/Research',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9'
        ],
        install_requires=[
            'pandas', 'IPython'
        ],
    )
